/*
 * fepg.c: An EPG plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 */

#include <vdr/plugin.h>
#include <vdr/interface.h>
#include <string>
#include <vector>
#include <sstream>
#include <stdlib.h>
#include <time.h>


using namespace std;

// Set to true if you are using Bell ExpressVu, false otherwise.
static const bool BEV = true;


static const char *VERSION        = "0.0.1";
static const char *DESCRIPTION    = "A Graphical EPG";
static const char *MAINMENUENTRY  = "fEPG";


string toTimeString(long);
string toTimeString(struct tm*);
int jumpTo(int sid, int length, char direction);

template<class type>
inline string to_string(const type& value)
{
  ostringstream streamOut;
  streamOut << value;
  return streamOut.str();
}
//////////////////////////////////////////////////////////////////////////////


class cFEpg : public cOsdObject 
{
private:
  void draw(int start_sid, char direction);
  void info(const cEventInfo*);
  void WrapText(string s, vector<string>&, int width, bool);


  cOsdBase *osd;

  static const int SCR_WIDTH  = 720; // Total screen width
  static const int SCR_HEIGHT = 480; // Total screen height
  static const int NAME_W = 125; // Width of the cells containing channel names & numbers
  static const int DX = 60;  // Space between edge of screen and display (horizontal)
  static const int DY = 20;  // Space between edge of screen and display (vertical)
  static const int DYO = 65; // Space between top of screen and the first channel display
  static const int NUM = 6;  // Number of channels shown at one time

  // Colors
  static const eDvbColor bgColor  = (eDvbColor) 0xFF687000;
  static const eDvbColor green    = (eDvbColor) 0xFF086E01;
  static const eDvbColor lightgrey= (eDvbColor) 0xFFAAAAAA;

  cBitmap* bitmap; // Do all drawing on this
  int width;  // Width of display
  int height; // Height of display (not counting the time bar at the top)
  int cWidth; // Width of a cell
  tWindowHandle windowHandle;
  
  int currentChannel; // The first channel shown in the EPG
  int displayedChannels[NUM]; // A list of the channels numbers currently shown 

  struct tm* firstTime; // The time shown in the first column of the EPG
  struct tm* currentTime;

  int curCellh, curCellv; // Currently selected Cell, horizontal and vertical values in EPG table
  const cEventInfo* curEvent; // Currently selected event

  bool info_open; // Is the info window open?
  bool wasMuted;  // Was the device muted when we entered the EPG

  const cEventInfo* eventMtx[NUM][3]; // Pointers to all the events currently shown on screen

  vector<int> inputNum; // Holds numerical inputs for channel selection
  long inputTime;       // The last time a numerical value was inputed


public:
  cFEpg(void);
  ~cFEpg();
  virtual void Show(void);
  virtual eOSState ProcessKey(eKeys Key);
 };

//----------------------------------------------------------------------------

cFEpg::cFEpg(void)
{
     osd = NULL;
     width = SCR_WIDTH - 2*DX;
     height = SCR_HEIGHT - DY - DYO;   
     cWidth = (width - NAME_W)/3;
     info_open = false;
     firstTime = new struct tm;
     curEvent = NULL;
     curCellh = 1;
     curCellv = 1;
     inputTime = 0;
}

//----------------------------------------------------------------------------

cFEpg::~cFEpg()
{
  if (osd != NULL)
    delete osd;
  if (bitmap != NULL)
    delete bitmap;
  if (firstTime != NULL)
    delete firstTime;
}

//----------------------------------------------------------------------------

void cFEpg::draw(int start_sid, char direction = 'D')
{
     time_t times[3];

     // Clear the event matrix
     for (int i=0; i<NUM; i++)
       for (int j=0; j<3; j++)
       {
         eventMtx[i][j] = NULL;
       }

     vector<const cSchedule*> ScheduleList;
     cMutexLock mutexLock;
     const cSchedules* Schedules = cSIProcessor::Schedules(mutexLock);

     // Check if channel is valid
     bool ok = false;
     while(!ok)
     {
       if (start_sid < 0) direction = 'D';
       if (start_sid > Channels.MaxNumber()) direction = 'U';
       cChannel* check = Channels.GetByNumber(start_sid);
       if (check)
         ok=true;
       else
       {
         if (direction == 'U')
           start_sid--;
         else
           start_sid++;
       }    
     }  
     
     currentChannel = start_sid;

     // Get current time
     time_t rawtime;
     time(&rawtime);
     currentTime = localtime(&rawtime);
/*
     // Update firstTime if necessary
     if ((mktime(firstTime) - mktime(currentTime))/60 <= -30)
     {
       // Add 0:30 to firstTime
       if (firstTime->tm_min == 30)
       { 
         firstTime->tm_min = 0;
         firstTime->tm_hour +=1;
         if (firstTime->tm_hour == 13)  firstTime->tm_hour = 1;
       }
       else
         firstTime->tm_min = 30;
     }
*/
     bitmap->Fill(0,0,720,480,clrBlack);
     bitmap->SetFont(fontOsd);
     bitmap->Fill(DX+NAME_W, DYO, DX+width, DYO+height, bgColor);
     bitmap->Fill(DX, DYO, DX+NAME_W, DYO+height, green);
     bitmap->Fill(DX, DY, DX+width, DYO, green);
     
     // Draw current time
     const char* s = toTimeString(currentTime).c_str();  
     bitmap->Text(DX+(NAME_W-(osd->Width(s)))/2, DY+10, s, lightgrey, green);


     // Draw the other times
     times[0] = mktime(firstTime);
     s = toTimeString(firstTime).c_str();
     bitmap->Text(DX+NAME_W+(cWidth-osd->Width(s))/2, DY+10, s, lightgrey, green);
     struct tm* t = new struct tm;
     (*t) = (*firstTime);

     if (firstTime->tm_min == 30)
     {
       t->tm_hour = firstTime->tm_hour + 1;
//       if (t->tm_hour == 13) t->tm_hour = 1;
       t->tm_min = 0;
       s = toTimeString(t).c_str();
       bitmap->Text(DX+NAME_W+cWidth+(cWidth-osd->Width(s))/2, DY+10, s, lightgrey, green);
       times[1] = mktime(t);

       t->tm_min = 30;
       s = toTimeString(t).c_str();
       bitmap->Text(DX+NAME_W+2*cWidth+(cWidth-osd->Width(s))/2, DY+10, s, lightgrey, green);
       times[2] = mktime(t); 
     }
     else
     {
       t->tm_hour = firstTime->tm_hour;
       t->tm_min = 30;
       s = toTimeString(t).c_str();
       bitmap->Text(DX+NAME_W+cWidth+(cWidth-osd->Width(s))/2, DY+10, s, lightgrey, green);
       times[1] = mktime(t);

       t->tm_hour = firstTime->tm_hour + 1;
//       if (t->tm_hour == 13) t->tm_hour = 1;
       t->tm_min = 0;
       s = toTimeString(t).c_str();
       bitmap->Text(DX+NAME_W+2*cWidth+(cWidth-osd->Width(s))/2, DY+10, s, lightgrey, green);
       times[2] = mktime(t);       
     }
     delete t;

     // Write the channel numbers & names
     int numberOfChans = 0;
     int n = 0;
     while(numberOfChans < NUM)
     {
        const char* chanNum = (to_string(start_sid + n)).c_str();
        cChannel* c = Channels.GetByNumber(start_sid + n);
        
        // Make sure this is a valid channel
        if (c)
        {
          displayedChannels[numberOfChans] = start_sid + n;
          const char* chanName = c->Name();
          bitmap->Text(DX+(NAME_W-(osd->Width(chanNum)))/2, DYO+5+numberOfChans*(height/NUM), chanNum, lightgrey, green);
          bitmap->Text(DX+(NAME_W-(osd->Width(chanName)))/2, DYO+30+numberOfChans*(height/NUM), chanName, lightgrey, green);
          numberOfChans++;

          // Get Schedule for this channel
          ScheduleList.push_back(Schedules->GetSchedule(c->GetChannelID()));
        }
           
        n++;
     }

     
     curEvent = NULL;
    

     //--- Put events in EPG -----
     for (int j=0; j<NUM; j++)
     {
       time_t stTime = times[0];
       long delTime = 3*30*60;
       while (stTime <= times[2])
       {
         int barX = width - NAME_W;  // place for the seperator: default value
         const cEventInfo* event = NULL;

         const char* title = "No Info";
         if (ScheduleList.at(j))
         {
           event = ScheduleList.at(j)->GetEventAround(stTime);
           if (event)
           {
             title = event->GetTitle();
             delTime = event->GetTime() + event->GetDuration() - stTime;
             // Calculate place for the seperator
             barX = (static_cast<float>(delTime/60))/30.0 * cWidth;
           }
         }
       
         int cellBuf = 0;


         if (stTime == times[0])
         {
           if (barX <= cWidth)
           {             
             barX = cWidth;
             eventMtx[j][0] = event;
           }
           else if (barX <= 2*cWidth)
           {
             barX = 2*cWidth;
             eventMtx[j][0] = event;
             eventMtx[j][1] = event;
           }
           else 
           {            
             barX = width - NAME_W;
             eventMtx[j][0] = event;
             eventMtx[j][1] = event;
             eventMtx[j][2] = event;
           }
         }
         else if (stTime == times[1])
         {
           cellBuf = cWidth;
           if (barX <= cWidth)
           {
             barX = cWidth; 
             eventMtx[j][1] = event;
           }
           else
           {
             barX = 2*cWidth;
             eventMtx[j][1] = event;
             eventMtx[j][2] = event;
           }
         }
         else if (stTime == times[2])
         {
           cellBuf = 2*cWidth;
           barX = cWidth;
           eventMtx[j][2] = event;
         }

         eDvbColor clr1 = lightgrey;
         eDvbColor clr2 = bgColor;

         // Check if this cell is selected
         if (eventMtx[j][curCellv-1] == event && j == curCellh-1) 
         {    
           curEvent = event;
           clr1 = bgColor;
           clr2 = lightgrey;
           bitmap->Fill(cellBuf+DX+NAME_W, DYO+j*(height/NUM), cellBuf+DX+NAME_W+barX, DYO+(j+1)*(height/NUM), lightgrey); //Fill
           bitmap->Fill(cellBuf+DX+NAME_W+3, DYO+j*(height/NUM), cellBuf+DX+NAME_W+4, DYO+(j+1)*(height/NUM), clrBlack);  //Left
           bitmap->Fill(cellBuf+DX+NAME_W+barX-1, DYO+j*(height/NUM), cellBuf+DX+NAME_W+barX+1, DYO+(j+1)*(height/NUM), clrBlack);  //Right
           bitmap->Fill(cellBuf+DX+NAME_W+3, DYO+j*(height/NUM)+3, cellBuf+DX+NAME_W+barX-1, DYO+j*(height/NUM)+4, clrBlack);   // Top
           bitmap->Fill(cellBuf+DX+NAME_W+3, DYO+(j+1)*(height/NUM)-1, cellBuf+DX+NAME_W+barX-1, DYO+(j+1)*(height/NUM), clrBlack);  // Bottom      
         }  

         // Draw seperator bar
         bitmap->Fill(cellBuf+DX+NAME_W+barX, DYO+j*(height/NUM), cellBuf+DX+NAME_W+barX+3, DYO+(j+1)*(height/NUM), lightgrey); 

         int availableWidth = barX;
         int offSet = 0;

         // Draw Arrow if event has already started or will not end
         if (event != NULL)
         {
           if (event->GetTime()+30*60 < times[0]) // DOESN'T WORK !!!
           {
             bitmap->Text(cellBuf+DX+NAME_W+6, DYO+20+j*(height/NUM), "<", clr1, clr2);
             offSet = osd->Width("<") - 3;
             availableWidth -= offSet;
           }
           else if (event->GetTime() + event->GetDuration() > (times[2]+60*30))
           {
             bitmap->Text(DX+width-15, DYO+20+j*(height/NUM), ">", clr1, clr2);
             availableWidth -= osd->Width(">");
           }    
         }    

         vector<string> wrappedTitle;
         string t(title);
         WrapText(t, wrappedTitle, availableWidth - 15, false);

         // Draw name of event
         if (wrappedTitle.size() == 1)
         {
           bitmap->Text(cellBuf+DX+NAME_W+offSet+(barX-osd->Width(wrappedTitle.at(0).c_str()))/2, DYO+20+j*(height/NUM), wrappedTitle.at(0).c_str(), clr1, clr2);
         }
         else if (wrappedTitle.at(0) == "" && wrappedTitle.size() > 1)
         {
           if (wrappedTitle.size() == 2)
             bitmap->Text(cellBuf+DX+NAME_W+offSet+(barX-osd->Width(wrappedTitle.at(1).c_str()))/2, DYO+20+j*(height/NUM), wrappedTitle.at(0).c_str(), clr1, clr2);
           else
           {
             bitmap->Text(cellBuf+DX+NAME_W+offSet+(barX-osd->Width(wrappedTitle.at(1).c_str()))/2, DYO+5+j*(height/NUM), wrappedTitle.at(0).c_str(), clr1, clr2);
             bitmap->Text(cellBuf+DX+NAME_W+offSet+(barX-osd->Width(wrappedTitle.at(2).c_str()))/2, DYO+30+j*(height/NUM), wrappedTitle.at(1).c_str(), clr1, clr2);
           }
         }
         else //TO DO: expand to cover more cases
         {
           bitmap->Text(cellBuf+DX+NAME_W+offSet+(barX-osd->Width(wrappedTitle.at(0).c_str()))/2, DYO+5+j*(height/NUM), wrappedTitle.at(0).c_str(), clr1, clr2);
           bitmap->Text(cellBuf+DX+NAME_W+offSet+(barX-osd->Width(wrappedTitle.at(1).c_str()))/2, DYO+30+j*(height/NUM), wrappedTitle.at(1).c_str(), clr1, clr2);
         }

         if (event)
         {
           stTime += delTime;

           if (stTime > times[0] && stTime <= times[1]) stTime = times[1];
           else if (stTime > times[1] && stTime <= times[2]) stTime = times[2];
           else if (stTime > times[2]) stTime = times[2] + 1000;
         }
         else // TO DO: What if there is no info now but there is some later
         {
           stTime = times[2] + 1000;
         }

       }
     }
      
     //---------------------------

   
     bitmap->Fill(DX, DY, DX+3, DYO+height+3, lightgrey); //Vert Left
     bitmap->Fill(DX+width, DY, DX+width+3, DYO+height+3, lightgrey); //Vert right

     bitmap->Fill(DX+NAME_W, DY, DX+NAME_W+3, DYO+height+3, lightgrey);
     
     // Vertical lines
     for (int i=1; i<3; i++)
     {
          bitmap->Fill(DX+NAME_W+i*cWidth, DY, DX+NAME_W+i*cWidth+3, DYO, lightgrey); 
     }

     
     bitmap->Fill(DX, DY, DX+width+3, DY+3, lightgrey);  // Top

     // Horizontal lines
     for (int k=0; k<=NUM; k++)
     {
          bitmap->Fill(DX, DYO+(height/NUM)*k, DX+width+3, DYO+(height/NUM)*k+3, lightgrey);
     }

     osd->SetBitmap(0,0,*bitmap, windowHandle);
     osd->Flush();
}

//----------------------------------------------------------------------------

void cFEpg::info(const cEventInfo* e)
{
  if (e == NULL)
    return;

  info_open = true;
  int W = (int)(0.8 * SCR_WIDTH);  // Width
  int H = (int)(0.7 * SCR_HEIGHT); // Height
  int B = 10; // Border
  
  int Y = (SCR_HEIGHT-H)/2; // Y Coord. of upper left corner
  int X = (SCR_WIDTH-W)/2;  // X Coord. of upper left corner
 
  // The Info Box
  bitmap->Fill(X, Y, X+W, Y+H, green);
  bitmap->Fill(X, Y, X+W, Y+2, clrBlack);     //top
  bitmap->Fill(X, Y+H-2, X+W, Y+H, clrBlack); //bottom
  bitmap->Fill(X, Y, X+2, Y+H, clrBlack);     //left
  bitmap->Fill(X+W-2, Y, X+W, Y+H, clrBlack); //right

  // The program name
  bitmap->Fill(X+B,Y+B, X+W-B, Y+50, lightgrey);
  bitmap->SetFont(fontOsd);
  const char* name = e->GetTitle() ? strdup(e->GetTitle()) : NULL;
  bitmap->Text(X+(W-(osd->Width(name)))/2, Y+15, name, clrBlack, lightgrey);

  // The description box
  int TOP = Y+2*B+50-10;
  bitmap->Fill(X+B, TOP, X+W-B, Y+H-B, lightgrey);
  bitmap->Fill(X+B, Y+H-B-60, X+W-B, Y+H-B-58, clrBlack);
  
  string start(e->GetTimeString());
  string end(e->GetEndTimeString());
  if (start.at(0) == '0')
    start.at(0) = ' ';
  if (end.at(0) == '0')
    end.at(0) = ' ';

  long lDuration = e->GetDuration();
  start = "Start  : " + start;
  end = "End   : " + end;
  string duration = "Duration : " + toTimeString(lDuration);
  int num = displayedChannels[curCellh-1];
  string channel = "Channel : " + to_string(num); 

  bitmap->Text(X+2*B, Y+H-B-55, start.c_str(), clrBlack, lightgrey);
  bitmap->Text(X+2*B, Y+H-B-30, end.c_str(), clrBlack, lightgrey);
  int side = osd->Width(duration.c_str());
  bitmap->Text(X+W-2*B-side, Y+H-B-55, duration.c_str(), clrBlack, lightgrey);
  bitmap->Text(X+W-2*B-side, Y+H-B-30, channel.c_str(), clrBlack, lightgrey);

  // The description text  
  vector<string> lines;
  string desc(e->GetExtendedDescription());
  WrapText(desc, lines, W-6*B, true);

  int buf, off;
  const char* subTitle;
  if (BEV)
  {
    subTitle = lines.at(0).c_str();
    off = 1;
  }
  else
  {
    subTitle = e->GetSubtitle();
    off = 0;
  }

  if (strcmp(subTitle, "") == 0)
    buf = BEV ? -25 : 0;
  else
  {
    buf = BEV ? 0 : 25;
    bitmap->Text(X+(W-osd->Width(subTitle))/2, TOP+10, subTitle, clrBlack, lightgrey);
  }

  //TO DO: use h to check that the text is contained in the box
  for (int n=off; n<lines.size(); n++)
  {
    bitmap->Text(X+2*B, TOP+10+buf+n*25, lines.at(n).c_str(), clrBlack, lightgrey); 
  }

  osd->SetBitmap(0,0,*bitmap, windowHandle);
  osd->Flush();

}

//----------------------------------------------------------------------------


void cFEpg::Show(void)
{

  // Mute
  if (!cDevice::PrimaryDevice()->IsMute())
  {
    wasMuted = false;
    cDevice::PrimaryDevice()->ToggleMute();
  }
  else
    wasMuted = true;
    
  osd = cOsd::OpenRaw(0, 0);

  if (osd) 
  {
     windowHandle = osd->Create(0, 0, SCR_WIDTH, SCR_HEIGHT, 2, false);
     osd->AddColor(clrBlack);
     osd->AddColor(bgColor);
     osd->AddColor(green);
     osd->AddColor(lightgrey);

     // Get current time
     time_t rawtime;
     time(&rawtime);
     currentTime = localtime(&rawtime);
     
     (*firstTime) = (*currentTime);
     //Round down firstTime to nearest 30 min
     if ((*firstTime).tm_min > 30)     
       (*firstTime).tm_min = 30;
     else
       (*firstTime).tm_min = 0;
     
     bitmap = new cBitmap(720,480,2,false);
     currentChannel = cDevice::CurrentChannel();
     draw(currentChannel);

  }
}

//----------------------------------------------------------------------------

eOSState cFEpg::ProcessKey(eKeys Key)
{
  eOSState state = cOsdObject::ProcessKey(Key);
  if (state == osUnknown) {
     switch (Key & ~k_Repeat) {

       // Move up by one channel
       case kUp:
         if (!info_open)
         {
           if (curCellh == 1)
             draw(--currentChannel, 'U');
           else
           {
             curCellh--;
             draw(currentChannel);
           }
         }
       break;
         
       // Move down by one channel
       case kDown:
         if (!info_open)
         {
           if (curCellh == NUM)
             draw(++currentChannel);
           else
           {
             curCellh++;
             draw(currentChannel);
           }
         }
       break;

       // Scroll left by 0:30
       case kLeft:    
         if (!info_open)
         {
           //Make sure firstTime doesn't get too small
           if ((mktime(firstTime) - mktime(currentTime))/60 <= 0 && curCellv == 1)
             break;

            if (curCellv == 1)
            {
              if (firstTime->tm_min == 30)
                firstTime->tm_min = 0;
              else
              {  
               firstTime->tm_hour -= 1;
               firstTime->tm_min = 30;
              // if (firstTime->tm_hour == 0) firstTime->tm_hour = 12;
              } 
            }
            else
            {
              bool stop = false;
              const cEventInfo* e = eventMtx[curCellh-1][curCellv-1];
              do
              { 
                if (curCellv == 1)
                {
                  ProcessKey(kLeft);
                  curCellv = 2;
                  stop = true;
                }
                curCellv--;
              } while(e == eventMtx[curCellh-1][curCellv-1] && !stop);
            }

             draw(currentChannel);
         }
       break;

       // Scroll right by 0:30
       case kRight:
         if (!info_open)
         {   
           if (curCellv == 3)
           {
             if (firstTime->tm_min == 30)
             { 
               firstTime->tm_min = 0;
               firstTime->tm_hour +=1;
               //if (firstTime->tm_hour == 13)  firstTime->tm_hour = 1;
             }
             else
               firstTime->tm_min = 30;
           }
           else
           {
              bool stop = false;
              const cEventInfo* e = eventMtx[curCellh-1][curCellv-1];
              do
              {
                if (curCellv == 3)
                {
                  ProcessKey(kRight);
                  curCellv = 2;
                  stop = true;
                }
                curCellv++;
              } while(e == eventMtx[curCellh-1][curCellv-1] && !stop);
              
           }
             draw(currentChannel);
         }
       break;

       // Open/close info window
       case kGreen:
         if (info_open)
         {
           info_open = false;
           draw(currentChannel);
         }
         else
         {
           info(curEvent);  
           info_open = true;
         }   
         break;

       // Page up
       case kRed:
         if (!info_open)
         {
           draw(jumpTo(currentChannel, NUM, 'U'));
         }
       break;

       // Page down
       case kYellow:
         if (!info_open)
         {
           draw(jumpTo(currentChannel, NUM, 'D'));
         }
       break;


       case kBlue:

       break;

       // Exit EPG, no channel switch
       case kBack:
           if (!wasMuted)
             cDevice::PrimaryDevice()->ToggleMute();
           return osEnd;
       break;
      
       //Go to channel#
       case k0 ... k9:
         if (inputTime == 0)  inputTime = time_ms();
         if (time_ms() - inputTime > 1000)  inputNum.clear();
 
         inputNum.push_back(Key - k0);
         inputTime = time_ms();
         
         // Does the number have 3 digits?
         if (inputNum.size() == 3)
         {
           curCellh = 1;
           draw(100*inputNum.at(0)+10*inputNum.at(1)+inputNum.at(2));
           inputNum.clear();
         }
       break;

       //Switch channel and exit or close info window
       case kOk:
         if (!info_open)
         {      
           // If the first event is selected, then switch
           if (eventMtx[curCellh-1][curCellv-1] == eventMtx[curCellh-1][0])
           {
             if (cDevice::CurrentChannel() != displayedChannels[curCellh-1])
             {
               bitmap->Fill(0,0,SCR_WIDTH,SCR_HEIGHT,clrBlack);
               osd->SetBitmap(0,0,*bitmap, windowHandle);
               osd->Flush();
               Channels.SwitchTo(displayedChannels[curCellh-1]);
             }
             if (!wasMuted)
               cDevice::PrimaryDevice()->ToggleMute();
             return osEnd;
           }
           else  // Open info window
           {
             info(curEvent);  
             info_open = true;
           }
         }
         else
         {
           info_open = false;
           draw(currentChannel);
         }
       break;
      
       default: return state;
       }

     
     state = osContinue;
     }
  return state;
}


//////////////////////////////////////////////////////////////////////////////


void cFEpg::WrapText(string s, vector<string>& lines, int width, bool title)
{
  // Tokenize
  vector<string> tokens;

  if (s == "")
  {
    lines.push_back("");
    return;
  }

  string word = "";
  for (int i = 0; i < s.length(); i++)
  {
    if (s.at(i) != ' ')
      word += s.at(i);
    else
    {
      tokens.push_back(word);
      word = "";
    }
  }

  tokens.push_back(word);


  int total = 0;
  string temp = "";
  bool foundTitle = false;
  for (int k = 0; k < tokens.size(); k++)
  {
    if (BEV && title && !foundTitle)
    {
      string st = tokens.at(k);
      if (st=="Series." || st=="Movie." || st=="News." || st=="Education." || st=="Sports.")
      {
        lines.push_back(temp);
        temp = "";
        total = 0;
        
        foundTitle = true;
      }
    }
    if (osd->Width(tokens.at(k).c_str()) + total >= width && temp != "")
    {
      lines.push_back(temp);
      temp = "";
      total = 0;
    }

    //Make sure the token is not too wide 
    while (osd->Width(tokens.at(k).c_str()) > width)
    {
      string split = "";
      int idx = 0;
      while (osd->Width(split.c_str()) <= width && idx < tokens.at(k).length())
      {
        split += tokens.at(k).at(idx++);
      }
      lines.push_back(split.substr(0, split.length()-2) + "-");
      tokens.at(k) = tokens.at(k).substr(idx-1);
    }

    temp += (tokens.at(k) + " ");
    total = osd->Width(temp.c_str());
  }
  
  if (temp != "")
    lines.push_back(temp);

}

//----------------------------------------------------------------------------

string toTimeString(long time)
{
  int mins = time / 60;
  int hrs  = mins / 60;
  mins = mins - hrs*60;
  string sMins = to_string(mins);
  if (sMins.length() == 1)
    sMins = "0" + sMins;

  return to_string(hrs) + ":" + sMins; 
}

//----------------------------------------------------------------------------

string toTimeString(struct tm* t)
{
  string mins = to_string((*t).tm_min);
  if (mins.length() == 1)
    mins = "0" + mins;
  int hrs = (*t).tm_hour;
  if (hrs > 12)
    hrs = hrs % 12;
  if (hrs == 0) hrs = 12;
  return to_string(hrs) + ":" + mins;
}

//----------------------------------------------------------------------------

// Has problems with high values
int jumpTo(int sid, int length, char direction)
{
  bool stop = false;
  for (int i=0; i<length && !stop; i++)
  {
     if (direction == 'U')
       sid--;
     else
       sid++;

     bool ok = false;
     while(!ok)
     {
       if (sid < 0)
       {
         direction = 'D';
         stop = true;
       }
       if (sid > 9999) 
       {
         direction = 'U';
         stop = true;
       }
       cChannel* check = Channels.GetByNumber(sid);
       if (check)
         ok=true;
       else
       {
         if (direction == 'U')
           sid--;
         else
           sid++;
       }    
     }  
   }
   return sid;
}


//////////////////////////////////////////////////////////////////////////////
// --- cPluginFepg --------------------------------------------------------

class cPluginFepg : public cPlugin {
private:
  // Add any member variables or functions you may need here.
public:
  cPluginFepg(void);
  virtual ~cPluginFepg();
  virtual const char *Version(void) { return VERSION; }
  virtual const char *Description(void) { return DESCRIPTION; }
  virtual const char *CommandLineHelp(void);
  virtual bool ProcessArgs(int argc, char *argv[]);
  virtual bool Start(void);
  virtual void Housekeeping(void);
  virtual const char *MainMenuEntry(void) { return MAINMENUENTRY; }
  virtual cOsdObject *MainMenuAction(void);
  virtual cMenuSetupPage *SetupMenu(void);
  virtual bool SetupParse(const char *Name, const char *Value);
  };

cPluginFepg::cPluginFepg(void)
{
  // Initialize any member variables here.
  // DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
  // VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!
}

cPluginFepg::~cPluginFepg()
{
  // Clean up after yourself!
}

const char *cPluginFepg::CommandLineHelp(void)
{
  // Return a string that describes all known command line options.
  return NULL;
}

bool cPluginFepg::ProcessArgs(int argc, char *argv[])
{
  // Implement command line argument processing here if applicable.
  return true;
}

bool cPluginFepg::Start(void)
{
  // Start any background activities the plugin shall perform.
  return true;
}

void cPluginFepg::Housekeeping(void)
{
  // Perform any cleanup or other regular tasks.
}

cOsdObject *cPluginFepg::MainMenuAction(void)
{
  // Perform the action when selected from the main VDR menu.
  return new cFEpg;
}

cMenuSetupPage *cPluginFepg::SetupMenu(void)
{
  // Return a setup menu in case the plugin supports one.
  return NULL;
}

bool cPluginFepg::SetupParse(const char *Name, const char *Value)
{
  // Parse your own setup parameters and store their values.
  return false;
}

VDRPLUGINCREATOR(cPluginFepg); // Don't touch this!
